@@ -161,9 +161,12 @@ class TestCallInspector(base.IronicAgentTest):
 
         res = inspector.call_inspector(data, failures)
 
-        mock_post.assert_called_once_with('url',
-                                          cert=None, verify=True,
-                                          data='{"data": 42, "error": null}')
+        mock_post.assert_called_once_with(
+            'url', data='{"data": 42, "error": null}',
+            cert=None, verify=True,
+            headers={'Content-Type': 'application/json',
+                     'Accept': 'application/json'},
+            timeout=30)
         self.assertEqual(mock_post.return_value.json.return_value, res)
 
     def test_send_failure(self, mock_post):
@@ -176,7 +179,9 @@ class TestCallInspector(base.IronicAgentTest):
 
         mock_post.assert_called_once_with('url',
                                           cert=None, verify=True,
-                                          data='{"data": 42, "error": "boom"}')
+                                          data='{"data": 42, "error": "boom"}',
+                                          headers=mock.ANY,
+                                          timeout=30)
         self.assertEqual(mock_post.return_value.json.return_value, res)
 
     def test_inspector_error(self, mock_post):
@@ -188,10 +193,13 @@ class TestCallInspector(base.IronicAgentTest):
 
         mock_post.assert_called_once_with('url',
                                           cert=None, verify=True,
-                                          data='{"data": 42, "error": null}')
+                                          data='{"data": 42, "error": null}',
+                                          headers=mock.ANY,
+                                          timeout=30)
         self.assertIsNone(res)
 
     @mock.patch.object(inspector, '_RETRY_WAIT', 0.01)
+    @mock.patch.object(inspector, '_RETRY_WAIT_MAX', 1)
     def test_inspector_retries(self, mock_post):
         mock_post.side_effect = requests.exceptions.ConnectionError
         failures = utils.AccumulatedFailures()
@@ -201,6 +209,38 @@ class TestCallInspector(base.IronicAgentTest):
                           data, failures)
         self.assertEqual(5, mock_post.call_count)
 
+    @mock.patch.object(inspector, '_RETRY_WAIT', 0.01)
+    @mock.patch.object(inspector, '_RETRY_WAIT_MAX', 1)
+    @mock.patch.object(inspector, '_RETRY_ATTEMPTS', 3)
+    def test_inspector_retries_on_50X_error(self, mock_post):
+        mock_post.side_effect = [mock.Mock(status_code=500),
+                                 mock.Mock(status_code=409),
+                                 mock.Mock(status_code=502)]
+        failures = utils.AccumulatedFailures()
+        data = collections.OrderedDict(data=42)
+        self.assertRaises(requests.exceptions.HTTPError,
+                          inspector.call_inspector,
+                          data, failures)
+        self.assertEqual(3, mock_post.call_count)
+
+    @mock.patch.object(inspector, '_RETRY_WAIT', 0.01)
+    @mock.patch.object(inspector, '_RETRY_WAIT_MAX', 1)
+    @mock.patch.object(inspector, '_RETRY_ATTEMPTS', 3)
+    def test_inspector_retry_on_50X_and_succeed(self, mock_post):
+        mock_post.side_effect = [mock.Mock(status_code=503),
+                                 mock.Mock(status_code=409),
+                                 mock.Mock(status_code=200)]
+
+        failures = utils.AccumulatedFailures()
+        data = collections.OrderedDict(data=42)
+        inspector.call_inspector(data, failures)
+        self.assertEqual(3, mock_post.call_count)
+        mock_post.assert_called_with('url',
+                                     cert=None, verify=True,
+                                     data='{"data": 42, "error": null}',
+                                     headers=mock.ANY,
+                                     timeout=30)
+
 
 class BaseDiscoverTest(base.IronicAgentTest):
     def setUp(self):
